"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createEntityStore = exports.createProducer = exports.trackEntity = exports.trackWrongValues = exports.getTargetTypes = exports.getInputTypes = exports.extractTargetTypes = exports.extractInputTypes = exports.produceEntries = exports.getProducerFields = exports.isProducerField = exports.trackUnknownProps = exports.mapObjectToEntity = exports.getStore = void 0;
const pick_1 = __importDefault(require("lodash/pick"));
const omit_1 = __importDefault(require("lodash/omit"));
const forIn_1 = __importDefault(require("lodash/forIn"));
const difference_1 = __importDefault(require("lodash/difference"));
const getDefaultStoreValues = (options) => {
    var _a;
    return ({
        unknown: {},
        mistyped: {},
        entities: {},
        meta: {
            enableWarnings: (_a = options === null || options === void 0 ? void 0 : options.enableWarnings) !== null && _a !== void 0 ? _a : false,
        },
    });
};
function getStore(options) {
    const store = globalThis.__ENTITY_OF__;
    function init() {
        if (!globalThis.__ENTITY_OF__) {
            globalThis.__ENTITY_OF__ = getDefaultStoreValues(options);
        }
    }
    function register(owner) {
        let store = Object.assign({}, globalThis.__ENTITY_OF__);
        if (!store.unknown[owner]) {
            store = Object.assign(Object.assign({}, store), { unknown: Object.assign(Object.assign({}, store.unknown), { [owner]: {} }) });
        }
        if (!store.mistyped[owner]) {
            store = Object.assign(Object.assign({}, store), { mistyped: Object.assign(Object.assign({}, store.mistyped), { [owner]: {} }) });
        }
        if (!store.entities[owner]) {
            store = Object.assign(Object.assign({}, store), { entities: Object.assign(Object.assign({}, store.entities), { [owner]: {} }) });
        }
        globalThis.__ENTITY_OF__ = store;
    }
    function set(type, owner, key) {
        return function (updater, initial) {
            let _s = globalThis.__ENTITY_OF__;
            _s[type][owner][key] = updater(_s[type][owner][key] || initial);
        };
    }
    function setEntity(owner) {
        return function (updater, initial) {
            let _s = globalThis.__ENTITY_OF__;
            _s.entities[owner] = updater(_s.entities[owner] || initial);
        };
    }
    function reset() {
        globalThis.__ENTITY_OF__ = getDefaultStoreValues(options);
    }
    function destroy() {
        delete globalThis.__ENTITY_OF__;
    }
    return {
        set,
        init,
        reset,
        store,
        destroy,
        register,
        setEntity,
    };
}
exports.getStore = getStore;
function mapObjectToEntity(data, Entity) {
    const fields = Entity.fields || [];
    const keys = fields.map(({ key }) => key);
    return Object.assign(new Entity(), (0, pick_1.default)(data, keys));
}
exports.mapObjectToEntity = mapObjectToEntity;
function trackUnknownProps(owner, inputKeys, targetKeys) {
    const keysDiff = (0, difference_1.default)(inputKeys, targetKeys);
    const { store, register, set } = getStore();
    const shouldWarn = store.meta.enableWarnings;
    if (keysDiff.length) {
        const key = JSON.stringify(keysDiff);
        const setUnknownCount = set("unknown", owner, key);
        if (!store[owner]) {
            register(owner);
        }
        setUnknownCount((count) => {
            if (shouldWarn && count === 0) {
                console.warn(`${owner} was initialized with unknown properties: ${keysDiff}`);
            }
            return count + 1;
        }, 0);
    }
}
exports.trackUnknownProps = trackUnknownProps;
function isProducerField({ key, type }, producerName) {
    const t = type();
    // Check first () => [Producer] types
    if (Array.isArray(t)) {
        if (t[0] === Object) {
            throw new Error(`${key} 'type' function should not return an array containing a primitive Object constructor`);
        }
        if (t[0] === Array) {
            throw new Error(`${key} 'type' function should not return an array containing a primitive Array constructor`);
        }
        if (typeof t[0] === "object" && t[0] !== null) {
            const target = Object.values(t[0])[0];
            switch (target) {
                case Object:
                    throw new Error(`${key} 'type' function should not return an array of records containing primitive Object constructors as values`);
                case Array:
                    throw new Error(`${key} 'type' function should not return an array of records containing primitive Array constructors as values`);
                case Number:
                    throw new Error(`${key} 'type' function should not return an array of records containing primitive Number constructors as values`);
                case String:
                    throw new Error(`${key} 'type' function should not return an array of records containing primitive String constructors as values`);
                case Boolean:
                    throw new Error(`${key} 'type' function should not return an array of records containing primitive Boolean constructors as values`);
                default:
                    return target.hasOwnProperty(producerName);
            }
        }
        return t[0].hasOwnProperty(producerName);
    }
    if (typeof t === "object" && !Array.isArray(t) && t !== null) {
        const target = Object.values(t);
        switch (target[0]) {
            case Object:
                throw new Error(`${key} 'type' function should not return a record containing a primitive Object constructor as value`);
            case Array:
                throw new Error(`${key} 'type' function should not return a record containing a primitive Array constructor as value`);
            default:
                return (target.filter((v) => v.hasOwnProperty(producerName)).length ===
                    target.length);
        }
    }
    switch (t) {
        case Array:
            throw new Error(`${key} 'type' function should not return a primitive Array constructor`);
        case Object:
            throw new Error(`${key} 'type' function should not return a primitive Object constructor`);
        case Number:
        case String:
        case Boolean:
            return false;
        default:
            if (!t.hasOwnProperty(producerName)) {
                throw new Error("Missing producer method - of");
            }
            return true;
    }
}
exports.isProducerField = isProducerField;
function getProducerFields(fields, producerName) {
    return fields.filter((field) => isProducerField(field, producerName));
}
exports.getProducerFields = getProducerFields;
function produceEntries(initialData, fieldsWithProducer) {
    const result = fieldsWithProducer.reduce((acc, { key, type, options }) => {
        if ((options === null || options === void 0 ? void 0 : options.optional) && typeof initialData[key] === "undefined") {
            return acc;
        }
        if ((options === null || options === void 0 ? void 0 : options.nullable) && initialData[key] === null) {
            return acc;
        }
        if (Array.isArray(type())) {
            if (typeof type()[0] === "object" &&
                type()[0] !== null) {
                return Object.assign(Object.assign({}, acc), { [key]: (initialData[key] || []).map((item) => {
                        return Object.entries(item).reduce((acc, curr) => {
                            const [k, v] = curr;
                            const entity = Object.values(type()[0])[0];
                            return Object.assign(Object.assign({}, acc), { [k]: entity.of(v) });
                        }, {});
                    }) });
            }
            return Object.assign(Object.assign({}, acc), { [key]: (initialData[key] || []).map((v) => {
                    if (options === null || options === void 0 ? void 0 : options.nullable)
                        return v;
                    return type()[0].of(v);
                }) });
        }
        if (typeof type() === "object" &&
            !Array.isArray(type()) &&
            type() !== null) {
            return Object.assign(Object.assign({}, acc), { [key]: Object.entries(initialData[key] || []).reduce((acc, curr) => {
                    const [k, v] = curr;
                    const entities = Object.values(type()).filter((v) => v.hasOwnProperty("of"));
                    if (options && (options === null || options === void 0 ? void 0 : options.producerFields)) {
                        if (options.producerFields[k]) {
                            const entity = entities.find((v) => { var _a; return v.name === ((_a = options === null || options === void 0 ? void 0 : options.producerFields) === null || _a === void 0 ? void 0 : _a[k]); });
                            if (entity) {
                                return Object.assign(Object.assign({}, acc), { [k]: entity.of(v) });
                            }
                        }
                        return Object.assign(Object.assign({}, acc), { [k]: v });
                    }
                    if (!entities[0])
                        return Object.assign(Object.assign({}, acc), { [k]: v });
                    return Object.assign(Object.assign({}, acc), { [k]: entities[0].of(v) });
                }, {}) });
        }
        return Object.assign(Object.assign({}, acc), { [key]: type().of(initialData[key]) });
    }, {});
    return result;
}
exports.produceEntries = produceEntries;
function extractInputTypes(fields) {
    return fields.reduce((acc, { key, options, value }) => {
        var _a, _b, _c;
        let out = Object.assign({}, acc);
        if (typeof value === "undefined") {
            if (options === null || options === void 0 ? void 0 : options.optional) {
                return out;
            }
        }
        if (value === null) {
            out[key] = `Null`;
            return out;
        }
        if (Array.isArray(value) && value.some((v) => v === null)) {
            out[key] = `Array<Null>`;
            return out;
        }
        let typeName = Array.isArray(value)
            ? value.length > 0
                ? (_b = (_a = value[0]) === null || _a === void 0 ? void 0 : _a.constructor) === null || _b === void 0 ? void 0 : _b.name
                : "Empty"
            : (_c = value === null || value === void 0 ? void 0 : value.constructor) === null || _c === void 0 ? void 0 : _c.name;
        if ((options === null || options === void 0 ? void 0 : options.isCustom) &&
            typeof value === "object" &&
            !Array.isArray(value) &&
            Object.values(value).length) {
            const values = Object.entries(value)
                .map(([k, v]) => {
                var _a;
                if ((_a = options === null || options === void 0 ? void 0 : options.producerFields) === null || _a === void 0 ? void 0 : _a[k]) {
                    return options.producerFields[k];
                }
                return v.constructor.name;
            })
                .sort();
            const uniqueValues = [...new Set(values)];
            typeName = `Record<${uniqueValues.join("|")}>`;
        }
        if (Array.isArray(value)) {
            out[key] = `Array<${typeName}>`;
            return out;
        }
        out[key] = `Primitive<${typeName}>`;
        return out;
    }, {});
}
exports.extractInputTypes = extractInputTypes;
function extractTargetTypes(fields) {
    const result = fields.reduce((acc, { key, type, options }) => {
        let out = Object.assign({}, acc);
        let typeName = "";
        if (Array.isArray(type())) {
            if (typeof type()[0] === "object" &&
                type()[0] !== null) {
                typeName = Object.values(type()[0])[0].name;
            }
            else {
                typeName = type()[0].name;
            }
        }
        else {
            if (typeof type() === "object" &&
                !Array.isArray(type()) &&
                type() !== null) {
                typeName = Object.values(type())
                    .map((v) => v.name)
                    .sort()
                    .join("|");
            }
            else {
                typeName = type().name;
            }
        }
        if (options === null || options === void 0 ? void 0 : options.nullable) {
            if (Array.isArray(type())) {
                if (typeof type()[0] === "object") {
                    out[key] = `NullableArray<Record<${typeName}>>`;
                    return out;
                }
                out[key] = `NullableArray<${typeName}>`;
                return out;
            }
            if (typeof type() === "object") {
                out[key] = `NullablePrimitive<Record<${typeName}>>`;
                return out;
            }
            out[key] = `NullablePrimitive<${typeName}>`;
            return out;
        }
        if (Array.isArray(type())) {
            if (typeof type()[0] === "object") {
                out[key] = `Array<Record<${typeName}>>`;
                return out;
            }
            out[key] = `Array<${typeName}>`;
            return out;
        }
        if (typeof type() === "object") {
            out[key] = `Primitive<Record<${typeName}>>`;
            return out;
        }
        out[key] = `Primitive<${typeName}>`;
        return out;
    }, {});
    return result;
}
exports.extractTargetTypes = extractTargetTypes;
function getInputTypes(fields, keysWithProducer) {
    const allInputTypes = extractInputTypes(fields);
    const primitiveInputTypes = (0, omit_1.default)(allInputTypes, keysWithProducer);
    return {
        allInputTypes,
        primitiveInputTypes,
    };
}
exports.getInputTypes = getInputTypes;
function getTargetTypes(fields, keysWithProducer) {
    const allTargetTypes = extractTargetTypes(fields);
    const primitiveTargetTypes = (0, omit_1.default)(allTargetTypes, keysWithProducer);
    return {
        allTargetTypes,
        primitiveTargetTypes,
    };
}
exports.getTargetTypes = getTargetTypes;
function trackWrongValues(owner, fields, keysWithProducer) {
    const { primitiveInputTypes } = getInputTypes(fields, keysWithProducer);
    const { allTargetTypes, primitiveTargetTypes } = getTargetTypes(fields, keysWithProducer);
    const { store, register, set, init } = getStore();
    const shouldWarn = store.meta.enableWarnings;
    init();
    if (!(store === null || store === void 0 ? void 0 : store.mistyped[owner])) {
        register(owner);
    }
    (0, forIn_1.default)(primitiveInputTypes, (inputType, key) => {
        const targetType = primitiveTargetTypes[key];
        const setMistypedCount = set("mistyped", owner, key);
        if (targetType.startsWith("NullableArray")) {
            if (inputType === "Null" ||
                inputType === "Array<Null>" ||
                inputType === "Array<Empty>") {
                return;
            }
            if (inputType.startsWith("Array") && targetType.includes(inputType)) {
                return;
            }
        }
        if (targetType.startsWith("NullablePrimitive<Record") &&
            targetType.includes("|")) {
            const targetTypes = targetType
                .replace("NullablePrimitive<Record<", "")
                .replace(">>", "")
                .split("|");
            const inputTypes = inputType
                .replace("Primitive<Record<", "")
                .replace(">>", "")
                .split("|");
            const isSubset = inputTypes.every((t) => targetTypes.includes(t));
            if (isSubset) {
                return;
            }
        }
        if (targetType.startsWith("NullablePrimitive")) {
            if (inputType === "Null" || inputType === "Array<Null>") {
                return;
            }
            if (targetType.includes(inputType)) {
                return;
            }
        }
        if (targetType.startsWith("Array") && inputType.includes("Empty")) {
            return;
        }
        if (targetType.startsWith("Primitive<Record") && targetType.includes("|")) {
            const targetTypes = targetType
                .replace("Primitive<Record<", "")
                .replace(">>", "")
                .split("|");
            const inputTypes = inputType
                .replace("Primitive<Record<", "")
                .replace(">>", "")
                .split("|");
            const isSubset = inputTypes.every((t) => targetTypes.includes(t));
            if (isSubset) {
                return;
            }
        }
        if (targetType !== inputType) {
            setMistypedCount((count) => {
                if (shouldWarn && count === 0) {
                    console.warn(`<${owner}.${key}: ${allTargetTypes[key]}> property received a mistyped value: ${inputType}`);
                }
                return count + 1;
            }, 0);
            return;
        }
    });
}
exports.trackWrongValues = trackWrongValues;
function trackEntity(name, fields) {
    const { store, setEntity, register } = getStore();
    if (!(store === null || store === void 0 ? void 0 : store.entities))
        return;
    if (!store.entities[name]) {
        register(name);
    }
    const { allTargetTypes } = getTargetTypes(fields, []);
    const prev = JSON.stringify(allTargetTypes);
    const next = JSON.stringify(getStore().store.entities[name]);
    if (prev === next)
        return;
    setEntity(name)(() => allTargetTypes);
}
exports.trackEntity = trackEntity;
function createProducer(target, producerName, callBack) {
    return function (data = {}) {
        const owner = target.name;
        const fields = target.fields;
        const targetKeys = fields.map(({ key }) => key);
        const inputKeys = Object.keys(data);
        const fieldsWithProducer = getProducerFields(fields, producerName);
        const keysWithProducer = fieldsWithProducer.map(({ key }) => key);
        const producedEntries = produceEntries(data, fieldsWithProducer);
        const producedData = Object.assign(Object.assign({}, data), producedEntries);
        const result = callBack
            ? callBack(producedData)
            : mapObjectToEntity(producedData, target);
        const fieldsWithValues = fields.map((field) => (Object.assign(Object.assign({}, field), { value: result[field.key] })));
        trackEntity(owner, fieldsWithValues);
        trackWrongValues(owner, fieldsWithValues, keysWithProducer);
        trackUnknownProps(owner, inputKeys, targetKeys);
        return result;
    };
}
exports.createProducer = createProducer;
function createEntityStore(options) {
    getStore(options).init();
}
exports.createEntityStore = createEntityStore;
