"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parse = void 0;
const logger_1 = require("./logger");
const utils_1 = require("./utils");
function parseArray(type, value, meta) {
    var _a;
    if (type.value.startsWith("Array")) {
        if (!Array.isArray(value)) {
            logger_1.Log.mismatch({
                expected: type,
                received: value,
                entityName: meta.owner,
                property: meta.key,
            });
            // valueIsNotArray
            return;
        }
        for (const val of value) {
            const targetTypes = Object.keys((_a = type.entities) !== null && _a !== void 0 ? _a : {});
            const valueType = (() => {
                if (typeof val === "object") {
                    if (val === null)
                        return "null";
                    if (Array.isArray(val))
                        return "Array";
                    return "Record";
                }
                return (0, utils_1.capitalize)(typeof val);
            })();
            if (targetTypes.includes(valueType)) {
                continue;
            }
            logger_1.Log.mismatch({
                expected: type,
                received: { overwrite: `Array<${valueType}>` },
                entityName: meta.owner,
                property: meta.key,
            });
        }
        const strippedType = type.value.replace("Array<", "").replace(">", "");
        const arrLength = value.length;
        for (let i = 0; i < arrLength; i++) {
            value[i] = parse(Object.assign(Object.assign({}, type), { value: strippedType }), value[i], meta);
        }
    }
    if (Array.isArray(value)) {
        return value.filter((v) => v !== undefined);
    }
    return value;
}
function parseRecord(type, value, meta) {
    if (type.value.startsWith("Record")) {
        if (typeof value !== "object" || Array.isArray(value) || value === null) {
            logger_1.Log.mismatch({
                expected: type,
                received: value,
                entityName: meta.owner,
                property: meta.key,
            });
            return "__mismatch__";
        }
        // check primitive record values for mismatch
        for (const key in value) {
            const splittedType = type.value.split(", ");
            const recordValueType = splittedType === null || splittedType === void 0 ? void 0 : splittedType[1];
            const inputType = (0, utils_1.capitalize)(typeof value[key]);
            if ((recordValueType === null || recordValueType === void 0 ? void 0 : recordValueType.includes(inputType)) ||
                !["String", "Number", "Boolean"].includes(inputType)) {
                continue;
            }
            logger_1.Log.mismatch({
                expected: type,
                received: { overwrite: `${splittedType[0]}, ${inputType}>` },
                entityName: meta.owner,
                property: meta.key,
            });
        }
        const strippedType = (() => {
            let arr = type.value.replace("Record<", "").replace(">", "").split(", ");
            arr.shift();
            return arr.join(", ");
        })();
        for (const key in value) {
            value[key] = parse(Object.assign(Object.assign({}, type), { value: strippedType }), value[key], meta);
        }
    }
}
function parseArrayUnion(t, type, value, meta) {
    if (Array.isArray(value)) {
        if (!t.startsWith("Array")) {
            logger_1.Log.mismatch({
                expected: type,
                received: value,
                entityName: meta.owner,
                property: meta.key,
            });
            return "__mismatch__";
        }
        const strippedType = t.replace("Array<", "").replace(">", "");
        const arrLength = value.length;
        for (let i = 0; i < arrLength; i++) {
            if (typeof value[i] !== "object") {
                continue;
            }
            value[i] = parse(Object.assign(Object.assign({}, type), { value: strippedType }), value[i], meta);
        }
        return value;
    }
}
function parseBasicUnion(t, type, value, meta) {
    var _a;
    if (!["String", "Number", "Boolean"].includes(t) &&
        ((_a = type === null || type === void 0 ? void 0 : type.entities) === null || _a === void 0 ? void 0 : _a[t]) &&
        typeof value === "object" &&
        !Array.isArray(value) &&
        value !== null) {
        return parse(Object.assign(Object.assign({}, type), { value: t }), value, meta);
    }
    logger_1.Log.mismatch({
        expected: type,
        received: value,
        entityName: meta.owner,
        property: meta.key,
    });
    return "__mismatch__";
}
function parseRecordOrEntityUnion(t, type, value, meta) {
    if (typeof value === "object" && !Array.isArray(value) && value !== null) {
        if (["String", "Number", "Boolean"].includes(t)) {
            logger_1.Log.mismatch({
                expected: type,
                received: value,
                entityName: meta.owner,
                property: meta.key,
            });
            return "__mismatch__";
        }
        const valueKeys = Object.keys(value);
        let valueIsEntity = false;
        for (const name in type.entities) {
            if (["String", "Number", "Boolean"].includes(name))
                continue;
            const entity = type.entities[name];
            const entityTypeFields = Object.keys(entity.types);
            if (entityTypeFields.length !== valueKeys.length) {
                valueIsEntity = false;
            }
            for (const key in value) {
                if (!entity.types[key]) {
                    valueIsEntity = false;
                    break;
                }
                valueIsEntity = true;
            }
            if (valueIsEntity) {
                return parse(Object.assign(Object.assign({}, type), { value: name }), value, meta);
            }
        }
        if (t.startsWith("Record") || !valueIsEntity) {
            const strippedType = (() => {
                if (!valueIsEntity) {
                    let recordType = "";
                    const types = type.value.split(" | ");
                    for (const name of types) {
                        if (!name.startsWith("Record"))
                            continue;
                        recordType = name;
                    }
                    return recordType;
                }
                let arr = t.replace("Record<", "").replace(">", "").split(", ");
                arr.shift();
                return arr.join(", ");
            })();
            // check primitive record values for mismatch
            for (const key in value) {
                const splittedType = strippedType.split(", ");
                const recordValueType = splittedType === null || splittedType === void 0 ? void 0 : splittedType[1];
                const inputType = (() => {
                    if (typeof value[key] === "object") {
                        if (value === null)
                            return "null";
                        if (Array.isArray(value[key])) {
                            return "Array";
                        }
                        return "Record";
                    }
                    return (0, utils_1.capitalize)(typeof value[key]);
                })();
                if ((recordValueType === null || recordValueType === void 0 ? void 0 : recordValueType.includes(inputType)) ||
                    ["String", "Number", "Boolean"].includes(inputType)) {
                    continue;
                }
                logger_1.Log.mismatch({
                    expected: type,
                    received: { overwrite: `${splittedType[0]}, ${inputType}>` },
                    entityName: meta.owner,
                    property: meta.key,
                });
                delete value[key];
            }
            if (Object.values(value).every((v) => typeof v === "object")) {
                for (const key in value) {
                    value[key] = parse(Object.assign(Object.assign({}, type), { value: strippedType }), value[key], meta);
                }
            }
            else {
                value = parse(Object.assign(Object.assign({}, type), { value: strippedType }), value, meta);
            }
            return value;
        }
    }
}
function parseUnion(type, value, meta) {
    if (type.value.includes(" | ") &&
        !type.value.startsWith("Record") &&
        !type.value.startsWith("Array")) {
        const types = type.value.split(" | ");
        const typesLength = types.length;
        let out;
        for (let i = 0; i < typesLength; i++) {
            const t = types[i];
            // if value is a non-object primitive
            if (typeof value !== "object" &&
                types.includes((0, utils_1.capitalize)(typeof value))) {
                break;
            }
            // if value is a record or an entity
            out = parseRecordOrEntityUnion(t, type, value, meta);
            if (out)
                return out;
            // if value is an array
            out = parseArrayUnion(t, type, value, meta);
            if (out)
                return out;
            // if value is either a non-object primitive or an entity
            out = parseBasicUnion(t, type, value, meta);
            if (out)
                return out;
        }
    }
}
function parsePrimitive(type, value, meta) {
    if ((typeof value === "object" &&
        ["String", "Number", "Boolean"].includes(type.value)) ||
        (typeof value === "string" && !type.value.includes("String")) ||
        (typeof value === "number" && !type.value.includes("Number")) ||
        (typeof value === "boolean" && !type.value.includes("Boolean"))) {
        logger_1.Log.mismatch({
            expected: type,
            received: value,
            entityName: meta.owner,
            property: meta.key,
        });
        return true;
    }
    return false;
}
function parse(type, value, meta) {
    let next = (0, utils_1.clone)(value);
    if (type.mapper && type.value === "custom")
        return type.mapper(next);
    if (type.isOptional && value === undefined)
        return;
    if (type.isNullable && value === null)
        return next;
    const array = parseArray(type, next, meta);
    if (array === undefined)
        return;
    const record = parseRecord(type, next, meta);
    if (record === "__mismatch__")
        return;
    const union = parseUnion(type, next, meta);
    if (union === "__mismatch__")
        return;
    if (union)
        next = union;
    if (type.entities) {
        const typeName = type.value;
        const entities = type.entities;
        const entity = entities[typeName];
        if (entity && Object.hasOwn(entity, "of")) {
            // parse record values
            if (typeof next === "object" && !Array.isArray(next) && next !== null) {
                if (Object.values(next).every((v) => typeof v !== "object")) {
                    next = entity.of(next);
                    return next;
                }
                let isMismatching = false;
                for (const key in next) {
                    if (["String", "Number", "Boolean"].includes(entity.types[key].value)) {
                        isMismatching = true;
                        logger_1.Log.mismatch({
                            expected: entity.types[key],
                            received: next[key],
                            entityName: typeName,
                            property: key,
                        });
                        next[key] = undefined;
                    }
                    else {
                        next[key] = entity.of(next[key]);
                    }
                }
                if (!isMismatching)
                    return next;
            }
            // parse singles entity values
            if (typeof next === "object" && !Array.isArray(next) && next !== null) {
                next = entity.of(next);
                return next;
            }
            else {
                logger_1.Log.mismatch({
                    expected: type,
                    received: value,
                    entityName: meta.owner,
                    property: meta.key,
                });
                return;
            }
        }
    }
    if (parsePrimitive(type, next, meta))
        return;
    if (Array.isArray(next)) {
        return next.filter((v) => v !== undefined);
    }
    if (typeof next === "object" && next !== null) {
        for (const key in next) {
            if (next[key] === undefined)
                delete next[key];
        }
    }
    return next;
}
exports.parse = parse;
